-- |A representation of cubes.

module Game.LupupuCube.Cube
    ( Cube(..)
    , CubeType(..)
    , Pattern(..)

    , isAlmighty
    , isFlush
    , isFlushable
    , isLinkedSomehow
    , hasSameTypeWith

    , ppCube
    )
    where

import           Control.Parallel.Strategies
import           Game.LupupuCube.XmlIO
import           Text.XML.HXT.Arrow

-- |A 'Cube' represents a cube.
data Cube
    = Cube {
        cuType         :: !CubeType -- ^The cube type.
      , cuIsDowny      :: !Bool     -- ^Does the cube float in midair?
      , cuIsFixed      :: !Bool     -- ^Is the cube unmovable from its first place?
      , cuIsLarge      :: !Bool     -- ^Is the cube large (2 x 2)?
      , cuHasUpperLink :: !Bool     -- ^Is the cube link to the upper one?
      , cuHasRightLink :: !Bool     -- ^Is the cube link to the right one?
      , cuHasLowerLink :: !Bool     -- ^Is the cube link to the lower one?
      , cuHasLeftLink  :: !Bool     -- ^Is the cube link to the left one?
      }
    deriving (Eq, Show, Ord)

instance NFData Cube where
    rnf c = rnf (cuType         c) >|
            rnf (cuIsDowny      c) >|
            rnf (cuIsFixed      c) >|
            rnf (cuIsLarge      c) >|
            rnf (cuHasUpperLink c) >|
            rnf (cuHasRightLink c) >|
            rnf (cuHasLowerLink c) >|
            rnf (cuHasLeftLink  c)

instance XmlRead Cube where
    xmlRead = proc e ->
              do ty  <- xmlRead  -< e
                 isD <- withDefault (getAttrValue0 "isDowny"     ) "no" >>> readBool -< e
                 isF <- withDefault (getAttrValue0 "isFixed"     ) "no" >>> readBool -< e
                 isL <- withDefault (getAttrValue0 "isLarge"     ) "no" >>> readBool -< e
                 up  <- withDefault (getAttrValue0 "hasUpperLink") "no" >>> readBool -< e
                 ri  <- withDefault (getAttrValue0 "hasRightLink") "no" >>> readBool -< e
                 lo  <- withDefault (getAttrValue0 "hasLowerLink") "no" >>> readBool -< e
                 le  <- withDefault (getAttrValue0 "hasLeftLink" ) "no" >>> readBool -< e
                 returnA -< Cube {
                               cuType         = ty
                             , cuIsDowny      = isD
                             , cuIsFixed      = isF
                             , cuIsLarge      = isL
                             , cuHasUpperLink = up
                             , cuHasRightLink = ri
                             , cuHasLowerLink = lo
                             , cuHasLeftLink  = le
                             }
        where
          readBool
              = proc t -> case t of
                            "yes" -> returnA -< True
                            "no"  -> returnA -< False
                            _     -> none    -< ()

instance XmlWrite Cube where
    xmlWrite = (arr cuType >>> xmlWrite)
               += (arr cuIsDowny      >>> omitDefault (attr "isDowny"      writeBool) False)
               += (arr cuIsFixed      >>> omitDefault (attr "isFixed"      writeBool) False)
               += (arr cuIsLarge      >>> omitDefault (attr "isLarge"      writeBool) False)
               += (arr cuHasUpperLink >>> omitDefault (attr "hasUpperLink" writeBool) False)
               += (arr cuHasRightLink >>> omitDefault (attr "hasRightLink" writeBool) False)
               += (arr cuHasLowerLink >>> omitDefault (attr "hasLowerLink" writeBool) False)
               += (arr cuHasLeftLink  >>> omitDefault (attr "hasLeftLink"  writeBool) False)
        where
          writeBool
              = proc b -> case b of
                            True  -> txt "yes" -< ()
                            False -> txt "no"  -< ()


omitDefault :: (ArrowList a, ArrowChoice a, Eq b) => a b c -> b -> a b c
omitDefault a def
    = proc b -> case b == def of
                  True  -> none -< ()
                  False -> a    -< b

-- |The type of cube.
data CubeType
    = Coloured !Pattern -- ^Coloured cube.
    | Almighty          -- ^Almighty rainbow cube.
    | Indelible         -- ^Indelible black cube.
    | Limpid            -- ^Limpid cube.
    | Flush             -- ^Flush bomb cube.
    deriving (Eq, Show, Ord)

instance NFData CubeType where
    rnf = rwhnf

instance XmlRead CubeType where
    xmlRead = proc e ->
              do name <- getLocalPart -< e
                 case name of
                   "coloured"
                       -> do s <- getAttrValue0 "kind" -< e
                             returnA -< Coloured (nameToPat s)
                   "almighty"
                       -> returnA -< Almighty
                   "indelible"
                       -> returnA -< Indelible
                   "limpid"
                       -> returnA -< Limpid
                   "flush"
                       -> returnA -< Flush
                   _
                       -> none    -< ()

instance XmlWrite CubeType where
    xmlWrite = proc ct ->
               case ct of
                 Coloured p
                     -> eelem "coloured" += attr "kind" mkText -< patToName p
                 Almighty
                     -> eelem "almighty"  -< ()
                 Indelible
                     -> eelem "indelible" -< ()
                 Limpid
                     -> eelem "limpid"    -< ()
                 Flush
                     -> eelem "flush"     -< ()

-- |Pattern of a coloured cube.
data Pattern
    = Pentagram
    | Hexagram
    | Square
    | Triangle
    | Diamond
    | Cross
    | Plus
    | Circle
    deriving (Eq, Ord, Show)


patToName :: Pattern -> String
patToName Pentagram = "pentagram"
patToName Hexagram  = "hexagram"
patToName Square    = "square"
patToName Triangle  = "triangle"
patToName Diamond   = "diamond"
patToName Cross     = "cross"
patToName Plus      = "plus"
patToName Circle    = "circle"


nameToPat :: String -> Pattern
nameToPat "pentagram" = Pentagram
nameToPat "hexagram"  = Hexagram
nameToPat "square"    = Square
nameToPat "triangle"  = Triangle
nameToPat "diamond"   = Diamond
nameToPat "cross"     = Cross
nameToPat "plus"      = Plus
nameToPat "circle"    = Circle
nameToPat name        = error ("undefined pattern: " ++ show name)


-- |Returns 'Prelude.True' iff the given 'Cube' is an 'Almighty' cube.
isAlmighty :: Cube -> Bool
isAlmighty c = case cuType c of
                 Almighty -> True
                 _        -> False

-- |Returns 'Prelude.True' iff the given 'Cube' is a 'Flush' cube.
isFlush :: Cube -> Bool
isFlush c = case cuType c of
              Flush -> True
              _     -> False


-- |Returns 'Prelude.True' iff the given 'Cube' can be flushed.
isFlushable :: Cube -> Bool
isFlushable c = case cuType c of
                  Coloured _ -> True
                  Almighty   -> True
                  Flush      -> True
                  _          -> False


-- |Returns 'Prelude.True' iff the given two 'Cube's has the same
-- type.
hasSameTypeWith :: Cube -> Cube -> Bool
hasSameTypeWith a b = cuType a == cuType b


-- |Returns a brief representation of the given 'Cube' in an ASCII
-- character.
ppCube :: Cube -> Char
ppCube c
    = case cuType c of
        Coloured Pentagram -> 'P'
        Coloured Hexagram  -> 'H'
        Coloured Square    -> 'S'
        Coloured Triangle  -> 'T'
        Coloured Diamond   -> 'D'
        Coloured Cross     -> 'C'
        Coloured Plus      -> 'L'
        Coloured Circle    -> 'I'
        Almighty  -> '%'
        Indelible -> '#'
        Limpid    -> ':'
        Flush     -> '!'

-- |Returns 'Prelude.True' iff the given 'Cube' is somehow linked to
-- another cube.
isLinkedSomehow :: Cube -> Bool
isLinkedSomehow c
    = cuHasUpperLink c ||
      cuHasRightLink c ||
      cuHasLowerLink c ||
      cuHasLeftLink  c
