{-# OPTIONS_HADDOCK prune #-}

-- |Representation of a situation in a gameplay.

module Game.LupupuCube.Phase
    ( Phase(..)
    , Position(..)
    , Size(..)
    , CubeMatrix

    -- private
    , Direction(..)

    -- private
    , walk
    , jump
    , fall
    , isCrushed
    , discreate
    , isCleared
    , isHopeless

    , ppPhase
    )
    where

import           Control.Monad
import           Control.Parallel.Strategies
import           Data.List
import qualified Data.Map as M
import           Data.Map (Map)
import           Data.Maybe
import qualified Data.Set as S
import           Data.Set (Set)
import           Game.LupupuCube.Cube
import           Game.LupupuCube.XmlIO
import           Prelude hiding (Left, Right)
import           Text.XML.HXT.Arrow hiding (left, right)

-- |'Phase' is a representation of a situation in a gameplay.
data Phase
    = Phase {
        pSaladPos :: !Position   -- ^The position of Salad.
      , pMatrix   :: !CubeMatrix -- ^The geometry of cubes and walls.
      }
    deriving (Eq, Show, Ord)

instance NFData Phase where
    rnf p = rnf (pMatrix   p) >|
            rnf (pSaladPos p)

instance XmlRead Phase where
    xmlRead = proc e ->
              do saladPos <- getSalad  >>> xmlRead -< e
                 matrix   <- getMatrix >>> readMat -< e
                 returnA -< Phase {
                               pSaladPos = saladPos
                             , pMatrix   = matrix
                             }
        where
          getMatrix = getChildren >>> isElem >>> hasName "matrix"
          getSalad  = getChildren >>> isElem >>> hasName "salad"
          readMat   = listA (getChildren >>> isElem >>> readCube) >>> arr M.fromList
          readCube  = xmlRead &&& xmlRead

instance XmlWrite Phase where
    xmlWrite = proc p ->
               do pos <- eelem "salad"  += xmlWrite -< pSaladPos p
                  mat <- eelem "matrix" += writeMat -< pMatrix   p
                  unlistA -< [pos, mat]
        where
          writeMat = arrL M.toList
                     >>>
                     ( (arr snd >>> xmlWrite)
                       +=
                       (arr fst >>> xmlWrite)
                     )

-- |'CubeMatrix' is a geometry of cubes and walls.
type CubeMatrix = Map Position Cube

type ForceSet = Set Position
    
-- |'Position' is an unpacked pair of two integers X and Y.
data Position
    = Pos {
        posX :: {-# UNPACK #-} !Int -- ^X
      , posY :: {-# UNPACK #-} !Int -- ^Y
      }
    deriving (Eq, Show, Ord)

instance NFData Position where
    rnf _ = ()

instance XmlRead Position where
    xmlRead = proc e ->
              do x <- getAttrValue0 "x" >>> arr read -< e
                 y <- getAttrValue0 "y" >>> arr read -< e
                 returnA -< Pos x y

instance XmlWrite Position where
    xmlWrite = proc (Pos x y) ->
               do x' <- attr "x" (arr show >>> mkText) -< x
                  y' <- attr "y" (arr show >>> mkText) -< y
                  unlistA -< [x', y']

-- |'Size' is an unpacked pair of two integers Width and Height.
data Size
    = Size {
        sWidth  :: {-# UNPACK #-} !Int -- ^Width
      , sHeight :: {-# UNPACK #-} !Int -- ^Height
      }
    deriving (Eq, Show, Ord)

instance NFData Size where
    rnf _ = ()

instance XmlRead Size where
    xmlRead = proc e ->
              do w <- getAttrValue0 "width"  >>> arr read -< e
                 h <- getAttrValue0 "height" >>> arr read -< e
                 returnA -< Size w h

instance XmlWrite Size where
    xmlWrite = proc (Size w h) ->
               do w' <- attr "width"  (arr show >>> mkText) -< w
                  h' <- attr "height" (arr show >>> mkText) -< h
                  unlistA -< [w', h']


data Direction
    = Left
    | Right
    | Up
    | Down
    deriving (Show, Eq)


movePos :: Position -> Direction -> Position
movePos (Pos x y) Left  = Pos (x - 1)  y
movePos (Pos x y) Right = Pos (x + 1)  y
movePos (Pos x y) Up    = Pos  x      (y + 1)
movePos (Pos x y) Down  = Pos  x      (y - 1)


isOutOfBound :: Size -> Position -> Bool
isOutOfBound (Size width height) (Pos x y)
    = case x < 0 of
        True  -> True
        False -> case y < 0 of
                   True  -> True
                   False -> case x >= width of
                              True  -> True
                              False -> y >= height


-- FIXME: 不完全。結合が考慮されてゐない。
mkForceSet :: Size -> Phase -> Position -> Direction -> ForceSet
mkForceSet size p pos dir = mkForceSet' size p pos dir S.empty

mkForceSet' :: Size -> Phase -> Position -> Direction -> ForceSet -> ForceSet
mkForceSet' size p pos dir soFar
    | isOutOfBound size pos
        = soFar

    | otherwise
        = case M.lookup pos (pMatrix p) of
            Nothing -> soFar
            Just _  -> mkForceSet' size p (movePos pos dir) dir (S.insert pos soFar)


-- ForceSet の全ての元が次の条件を満たす場合に True
--
-- 1. そこにあるオブジェクトそのものが空間に固定されてゐたら False
--
-- 2. 移動先が空間の境界外だったら False
--
-- 3. 移動先と等しい元が ForceSet に存在せず、且つ同じ場所にキューブが
--    あったら False
-- 
-- 4. いずれも成り立たなければ True
--
isEffective :: Size -> Phase -> Direction -> ForceSet -> Bool
isEffective size p dir fs = all isEffective' (S.toList fs)
    where
      isEffective' :: Position -> Bool
      isEffective' pos
          | fmap isFixed (getCubeAt p pos) == Just True = False
          | isOutOfBound size (movePos pos dir)           = False
          | (movePos pos dir `S.notMember` fs) &&
            (movePos pos dir `M.member` pMatrix p)        = False
          | otherwise                                     = True

      isFixed :: Cube -> Bool
      isFixed cube
          | cuIsFixed cube                = True
          | cuIsDowny cube && dir == Down = True
          | otherwise                     = False


getCubeAt :: Phase -> Position -> Maybe Cube
getCubeAt p pos = case M.lookup pos $! pMatrix p of
                    m -> m `seq` m


-- 1. 移動先が境界外なら False
--
-- 2. 移動先にキューブが存在しないか、存在したとしてもそのキューブが移
--    動しようとしてゐるなら True
--
-- 3. さうでなければ False
canSaladMoveTo :: Size -> Phase -> Direction -> ForceSet -> Bool
canSaladMoveTo size p dir fs
    | isOutOfBound size pos'          = False
    | isNothing (getCubeAt p pos')    = True
    | (isEffective size p dir fs) &&
      (pos' `S.member` fs)            = True
    | otherwise                       = False
    where
      pos' :: Position
      pos' = movePos (pSaladPos p) dir


-- 1. 進行方向に一つ進んだ位置を起點にして ForceSet を作る。
--
-- 2. その ForceSet を使って、同じ位置へサラダが移動可能かどうかを調べ
--    る。
--
-- 3. 移動可能な場合には、ForceSet にサラダ自身を追加して applyForce す
--    る。
--
walk :: Size -> Phase -> Direction -> Maybe Phase
walk size p dir
    | dir == Left || dir == Right
        = let next = movePos (pSaladPos p) dir
              fs   = mkForceSet size p next dir
              fs'  = S.insert (pSaladPos p) fs
          in
            if canSaladMoveTo size p dir fs then
                Just $ applyForce p dir fs'
            else
                Nothing
    | otherwise
        = error ("Can't walk " ++ show dir)


-- 1. 上は空欄か？空欄なら上へ移動。空欄でなければエラー終了。
--
-- 2. 右/左へ歩けるか？歩けるなら歩いて Phase を更新。歩けなければエラー
--    終了。
--
-- 3. 下は空欄か？空欄でなければ正常終了。空欄ならば次へ。
--
-- 4. 右/左は空欄か？空欄ならば右/左へ歩いて Phase を更新して正常終了。
--    空欄でなければその場で正常終了。
--
jump :: Size -> Phase -> Direction -> Maybe Phase
jump size p dir
    | dir == Left || dir == Right
        = do pUp     <- if canSaladMoveTo size p Up S.empty then
                            Just $ applyForce p Up $ S.singleton $ pSaladPos p
                        else
                            Nothing
             pHoriz1 <- walk size pUp dir
             return $ if canSaladMoveTo size pHoriz1 Down S.empty then
                          if canSaladMoveTo size pHoriz1 dir S.empty then
                              applyForce pHoriz1 dir $ S.singleton $ pSaladPos pHoriz1
                          else
                              pHoriz1
                      else
                          pHoriz1
    | otherwise
        = error ("Can't jump " ++ show dir)


-- 1. 全てのオブジェクトに對し、その位置を起點に作った ForceSet の集合
--    を得る。
--
-- 2. その集合の元のうち有效なものだけを集めて合成する。
--
-- 3. もしサラダが下へ移動可能なら、サラダ自身を ForceSet に追加する。
--
-- 4. 得られた ForceSet を下向きに適用する。
--
fall :: Size -> Phase -> Maybe Phase
fall size p
    = let forceSets   = do origins <- pSaladPos p : M.keys (pMatrix p)
                           return (mkForceSet size p origins Down)
                        `using`
                        parList rnf
          filteredFSs = filter (isEffective size p Down) forceSets
                        `using`
                        parList rwhnf
          mergedFS    = S.unions filteredFSs
          mergedFS'   = if canSaladMoveTo size p Down mergedFS then
                            S.insert (pSaladPos p) mergedFS
                        else
                            mergedFS
      in
        if S.null mergedFS' then
            Nothing
        else
            Just $ applyForce p Down mergedFS'


isCrushed :: Phase -> Bool
isCrushed p = pSaladPos p `M.member` pMatrix p


discreate :: Phase -> Maybe Phase
discreate p
    | p' == p   = Nothing
    | otherwise = Just p'
    where
      p' :: Phase
      p' = p { pMatrix = discreate' (pMatrix p) }

      discreate' :: CubeMatrix -> CubeMatrix
      discreate' = join scanCubes

      scanCubes :: CubeMatrix -> CubeMatrix -> CubeMatrix
      scanCubes tmp result
          = case M.minViewWithKey tmp of
              Nothing
                  -> result
              Just ((pos, cube), tmp')
                  -> case cuType cube of
                       ct@(Coloured _)
                           -> let cubes = findAdjacent tmp ct pos S.empty
                                  tmp'' = foldl' (flip M.delete) tmp (S.toList cubes)
                              in
                                if S.size cubes >= 3 then
                                    let result' = foldl' (flip M.delete) result (S.toList cubes)
                                    in
                                      scanCubes tmp'' result'
                                else
                                    scanCubes tmp'' result
                       Flush
                           -> let cubes      = neighborhoods result pos
                                  flushables = filter isFlushable cubes
                              in
                                if null flushables then
                                    scanCubes tmp' result
                                else
                                    let f c     = not $ any ($ c) [ hasSameTypeWith c' | c' <- flushables ]
                                        tmp''   = M.filter f tmp'
                                        result' = M.delete pos $ M.filter f result
                                    in
                                      scanCubes tmp'' result'
                       _   -> scanCubes tmp' result


isCleared :: Phase -> Bool
isCleared p = scanCubes (M.elems (pMatrix p))
    where
      scanCubes :: [Cube] -> Bool
      scanCubes []     = True
      scanCubes (x:xs) = case cuType x of
                           Indelible -> scanCubes xs
                           _         -> False


hasAlmighty :: Phase -> Bool
hasAlmighty p = any isAlmighty $ M.elems $ pMatrix p


hasFlush :: Phase -> Bool
hasFlush p = any isFlush $ M.elems $ pMatrix p


isHopeless :: Phase -> Bool
isHopeless p
    | hasAlmighty p      = False
    | hasFlush p         = False
    | notEnoughCubes     = True
    | otherwise          = False
    where
      notEnoughCubes :: Bool
      notEnoughCubes = {-# SCC "notEnoughCubes" #-}
                       any (< 3) (M.elems colouredMap)

      colouredMap :: Map Pattern Int
      colouredMap = {-# SCC "colouredMap" #-}
                    mkColouredMap $ pMatrix p


mkColouredMap :: CubeMatrix -> Map Pattern Int
mkColouredMap matrix = mkMap' M.empty $ M.elems matrix
    where
      mkMap' :: Map Pattern Int -> [Cube] -> Map Pattern Int
      mkMap' m []     = m
      mkMap' m (x:xs) = case cuType x of
                          Coloured code
                              -> let altered = M.alter inc code m
                                 in
                                   mkMap' altered xs
                          _   -> mkMap' m xs

      inc :: Maybe Int -> Maybe Int
      inc Nothing  = Just 1
      inc (Just n) = Just (n + 1)


findAdjacent :: CubeMatrix -> CubeType -> Position -> Set Position -> Set Position
findAdjacent m ct pos soFar
    | pos `S.member` soFar = soFar
    | otherwise
        = case M.lookup pos m of
            Nothing -> soFar
            Just cube
                | cuType cube == ct
                    -> findAdjacent m ct (movePos pos Up   ) $|| rnf $
                       findAdjacent m ct (movePos pos Right) $|| rnf $
                       findAdjacent m ct (movePos pos Down ) $|| rnf $
                       findAdjacent m ct (movePos pos Left ) $|| rnf $
                       S.insert pos soFar
                | otherwise
                    -> soFar


neighborhoods :: CubeMatrix -> Position -> [Cube]
neighborhoods m pos = catMaybes $ map nei' [Up, Right, Down, Left]
    where
      nei' :: Direction -> Maybe Cube
      nei' dir = M.lookup (movePos pos dir) m


applyForce :: Phase -> Direction -> ForceSet -> Phase
applyForce p dir fs = apply' p forcedPoints
    where
      forcedPoints :: [Position]
      forcedPoints = sortBy (comparator dir) (S.toList fs)

      apply' :: Phase -> [Position] -> Phase
      apply' p' []         = p'
      apply' p' (pos:rest) = if pos == pSaladPos p' then
                                 p' { pSaladPos = movePos pos dir }
                             else
                                 let Just cube = M.lookup pos (pMatrix p')
                                     deleted   = M.delete pos (pMatrix p')
                                     inserted  = M.insert (movePos pos dir) cube deleted
                                     p''       = p' { pMatrix = inserted }
                                 in
                                   apply' p'' rest


comparator :: Direction -> Position -> Position -> Ordering
comparator Left  = leftToRight
comparator Right = rightToLeft
comparator Up    = higherToLower
comparator Down  = lowerToHigher


-- 比較函數。x 座標が小さいほど前に來る。
leftToRight :: Position -> Position -> Ordering
leftToRight (Pos x1 _) (Pos x2 _)
    | x1 <  x2  = LT
    | x1 == x2  = EQ
    | otherwise = GT


-- 比較函數。x 座標が大きいほど前に來る。
rightToLeft :: Position -> Position -> Ordering
rightToLeft (Pos x1 _) (Pos x2 _)
    | x1 <  x2  = GT
    | x1 == x2  = EQ
    | otherwise = LT


-- 比較函數。y 座標が小さいほど前に來る。
lowerToHigher :: Position -> Position -> Ordering
lowerToHigher (Pos _ y1) (Pos _ y2)
    | y1 <  y2  = LT
    | y1 == y2  = EQ
    | otherwise = GT


-- 比較函數。y 座標が大きいほど前に來る。
higherToLower :: Position -> Position -> Ordering
higherToLower (Pos _ y1) (Pos _ y2)
    | y1 <  y2  = GT
    | y1 == y2  = EQ
    | otherwise = LT


-- |Returns a brief representation of the given 'Phase' in an ASCII
-- art. 'Size' is also needed because 'Phase' doesn't know anything
-- about the size of stage.
ppPhase :: Size -> Phase -> String
ppPhase size p = unlines' rows
    where
      unlines' :: [String] -> String
      unlines' []       = ""
      unlines' [x]      = x
      unlines' (x:y:zs) = x ++ "\n" ++ unlines' (y:zs)

      rows :: [String]
      rows = do y <- reverse [0 .. sHeight size - 1]
                return (row y)

      row :: Int -> String
      row y = do x <- [0 .. sWidth size - 1]
                 return ( case M.lookup (Pos x y) (pMatrix p) of
                            Nothing   -> if (Pos x y) == pSaladPos p then
                                             '@'
                                         else
                                             '.'
                            Just cube -> ppCube cube
                        )

