-- |The engine of the Lupupu Cube solver.

module Game.LupupuCube.Solver
    ( LupuGraph(..)
    , Depth
    , Pending(..)
    , FinalState(..)
    , Solution
    , Action(..)

    , isFinished
    , numOfPendings
    , numOfStates
    , nextPending

    , solveM
    , solve

    , ppAction
    )
    where

import           Control.Monad.Identity
import qualified Data.Graph.Inductive.Graph as G
import           Data.Graph.Inductive.Query.SP
import           Data.List
import           Data.Map (Map)
import qualified Data.Map as M
import           Data.Set (Set)
import qualified Data.Set as S
import           Game.LupupuCube.Graph
import           Game.LupupuCube.Phase
import           Game.LupupuCube.Stage
import           Prelude hiding (Left, Right)

-- |'Depth' is just an 'Prelude.Int' which represents a depth of node
-- in a 'LupuGraph'.
type Depth    = Int

-- |'Solution' is a list of tuple @('Action', 'Phase')@, representing
-- the shortest path from the initial state to one of the final
-- states.
type Solution = [(Action, Phase)]

-- |'LupuGraph' is a data structure which represents the intermediate
-- state of stage solving.
data LupuGraph
    = LupuGraph {
        -- |The size of stage.
        lgSize         :: !Size

      -- |The directed graph of state of the stage, which is usually
      -- incomplete (i.e. doesn't have every possible states).
      , lgGraph        :: !(Gr Depth Action)

      -- |The finite map from 'Phase' to 'G.Node'.
      , lgPhaseMap     :: !(Map Phase G.Node)

      -- |The finite set of states which are known to be possible but
      -- not actually examined yet. At the time this set becomes
      -- empty, the stage is either solved or proven to have no
      -- solutions.
      , lgPending      :: !(Set Pending)

      -- |The best solution ever found, which has the shortest path to
      -- one of final states.
      , lgBestSolution :: !(Maybe FinalState)

      -- |The 'G.Node' of the initial state.
      , lgInitialNode  :: !G.Node

      -- |The initial state.
      , lgInitialPhase :: !Phase
      }

-- |'Pending' represents a state which is known to be possible but not
-- actually determined yet. See 'lgPending'.
data Pending
    -- |The 'G.Node' of the state, the number of moves from the
    -- initial state to this state, and the state itself.
    = PEN
      {-# UNPACK #-} !G.Node 
      {-# UNPACK #-} !Depth
      !Phase
    deriving (Eq)

-- |'FinalState' represents a final state.
data FinalState
    -- |The 'G.Node' of the state, and the number of moves from the
    -- initial state to this state.
    = FST
      {-# UNPACK #-} !G.Node
      {-# UNPACK #-} !Depth
               

instance Ord Pending where
    (PEN n1 d1 p1) `compare` (PEN n2 d2 p2)
        = case d1 `compare` d2 of
            LT -> LT
            GT -> GT
            EQ -> case n1 `compare` n2 of
                    LT -> LT
                    GT -> GT
                    EQ -> p1 `compare` p2

-- |'Action' is a data type to represent a way of state transition.
data Action = WalkLeft  -- ^Walk one step to left.
            | WalkRight -- ^Walk one step to right.
            | JumpLeft  -- ^Jump one step to left.
            | JumpRight -- ^Jump one step to right.
              deriving (Eq, Show)

-- |Returns 'Prelude.True' iff the given 'LupuGraph' has no pending
-- states.
isFinished :: LupuGraph -> Bool
isFinished = S.null . lgPending

-- |Returns the number of pending states of the given 'LupuGraph'.
numOfPendings :: LupuGraph -> Int
numOfPendings = S.size . lgPending

-- |Returns the number of examined states of the given 'LupuGraph'.
numOfStates :: LupuGraph -> Int
numOfStates = G.noNodes . lgGraph

-- |Returns the next pending state of the given 'LupuGraph' if any.
nextPending :: LupuGraph -> Maybe Pending
nextPending lg
    | S.null (lgPending lg) = Nothing
    | otherwise             = Just (S.findMin (lgPending lg))


lookupNode :: Phase -> LupuGraph -> Maybe G.Node
lookupNode = (. lgPhaseMap) . M.lookup


insPhase :: Phase -> Depth -> LupuGraph -> (# G.Node, LupuGraph #)
insPhase p d lg
    = let [node] = G.newNodes 1 (lgGraph lg)
          graph' = G.insNode (node, d) (lgGraph lg)
          pm'    = M.insert p node (lgPhaseMap lg)
          lg'    = lg {
                     lgGraph    = graph'
                   , lgPhaseMap = pm'
                   }
      in
        (# node, lg' #)


insAction :: G.Node -> G.Node -> Action -> LupuGraph -> LupuGraph
insAction a b action lg = lg'
    where
      ledge  = (a, b, action)
      graph' = G.insEdge ledge (lgGraph lg)
      lg'    = lg { lgGraph = graph' }


extend :: LupuGraph -> LupuGraph
extend lg
    | isFinished lg = lg
    | otherwise
        = case S.deleteFindMin (lgPending lg) of
            (x@(PEN _ depth _), xs)
                -> case lgBestSolution lg of
                     Just (FST _ d) | d < depth -> lg { lgPending = S.empty } -- cut off
                     _ -> extend' x WalkLeft  $!
                          extend' x WalkRight $!
                          extend' x JumpLeft  $!
                          extend' x JumpRight $!
                          lg { lgPending = xs }
    where
      size :: Size
      size = lgSize lg

      extend' :: Pending -> Action -> LupuGraph -> LupuGraph
      extend' (PEN node depth p) action g'
          = case tryStateChange size p action of
              Nothing -> g'
              Just p' -> case lookupNode p' g' of
                           Nothing
                               -> let depth'           = depth + 1
                                      (# node', g'' #) = insPhase p' depth' g'
                                      g'''             = insAction node node' action g''
                                      g'''' | isCleared p'  = g''' {
                                                                lgBestSolution = Just (FST node' depth')
                                                              }
                                            | isHopeless p' = g'''
                                            | otherwise     = g''' {
                                                                lgPending = S.insert (PEN node' depth' p') (lgPending g''')
                                                              }
                                  in
                                    g''''
                           Just node'
                               -> insAction node node' action g'


mkInitialGraph :: Stage -> LupuGraph
mkInitialGraph st
    = let emptyGraph    = G.empty
          [initialNode] = G.newNodes 1 emptyGraph
          initialP      = stInitial st
          lnode         = (initialNode, 0)
          initialGraph  = G.insNode lnode emptyGraph
          initialPM     = M.singleton initialP initialNode
          initialQ      = S.singleton (PEN initialNode 0 initialP)
          initialLG     = LupuGraph {
                            lgSize         = stSize st
                          , lgGraph        = initialGraph
                          , lgPhaseMap     = initialPM
                          , lgPending      = initialQ
                          , lgBestSolution = Nothing
                          , lgInitialNode  = initialNode
                          , lgInitialPhase = initialP
                          }
      in
        initialLG


mkGraphM :: forall m a. Monad m => Stage -> (LupuGraph -> m a) -> m LupuGraph
mkGraphM st f = fullyExtend $ mkInitialGraph st
    where
      fullyExtend :: LupuGraph -> m LupuGraph
      fullyExtend lg
          | isFinished lg = f lg >> return lg
          | otherwise     = f lg >> fullyExtend (extend lg)


-- |Solves a 'Stage' in a 'Prelude.Monad' with repeatedly applying an
-- monadic action to the intermediate 'LupuGraph'. The 'Prelude.Monad'
-- is usually either 'Prelude.IO' or
-- 'Control.Monad.Identity.Identity'. See also 'solve'.
solveM :: Monad m =>
          Stage              -- ^The 'Stage' to solve.
       -> (LupuGraph -> m a) -- ^An action to be called back repeatedly during the stage solving.
       -> m (Maybe Solution) -- ^The solution if found.
solveM st f
    = do lg <- mkGraphM st f
         return $ do FST final _ <- lgBestSolution lg
                     let nodePath   = sp (lgInitialNode lg) final (weightGraph $ lgGraph lg)
                         actionPath = wriggle (lgGraph lg) nodePath
                         solution   = applyActions (lgInitialPhase lg) actionPath
                     return solution
    where
      size :: Size
      size = stSize st

      weightGraph :: Gr a Action -> Gr a Double
      weightGraph g = G.emap actionToWeight g

      applyActions :: Phase -> [Action] -> [(Action, Phase)]
      applyActions _ []     = []
      applyActions p (x:xs) = let Just p' = tryStateChange size p x
                              in
                                (x, p') : applyActions p' xs
{-# SPECIALISE solveM :: Stage -> (LupuGraph -> IO a) -> IO (Maybe Solution) #-}

-- |Solves a 'Stage' in an 'Control.Monad.Identity.Identity' monad
-- with no-op callback. See also 'solveM'.
solve :: Stage -> Maybe Solution
solve = runIdentity . flip solveM (return . const ())


wriggle :: G.Graph gr => gr a b -> [G.Node] -> [b]
wriggle _ []       = []
wriggle _ (_:[])   = []
wriggle g (x:y:zs) = b : wriggle g (y:zs)
    where
      Just b = findSuccEdge g x y


findSuccEdge :: G.Graph gr => gr a b -> G.Node -> G.Node -> Maybe b
findSuccEdge g x y
    = fmap snd $ find ((== y) . fst) $ G.lsuc g x


-- 1. 移動またはジャンプを試行。失敗したらエラー終了。
--
-- 2. 落下できる物が何も無くなるまで落下。
--
-- 3. 押し潰されたかどうかを判定する。押し潰されたらエラー終了。
--
-- 4. 消せる物を消す。もし消せる物が無かったら正常終了。
--
-- 5. 2 へ戻る。
tryStateChange :: Size -> Phase -> Action -> Maybe Phase
tryStateChange size p action = tryAction
    where
      tryAction :: Maybe Phase
      tryAction = do p' <- case action of
                             WalkLeft  -> walk size p Left
                             WalkRight -> walk size p Right
                             JumpLeft  -> jump size p Left
                             JumpRight -> jump size p Right
                     tryFall p'
                    
      tryFall :: Phase -> Maybe Phase
      tryFall p' = case fall size p' of
                     Just p'' -> tryFall p''
                     Nothing  -> tryCrush p'

      tryCrush :: Phase -> Maybe Phase
      tryCrush p'
          | isCrushed p' = Nothing
          | otherwise    = tryDiscreate p'

      tryDiscreate :: Phase -> Maybe Phase
      tryDiscreate p'
          = case discreate p' of
              Just p'' -> tryFall p''
              Nothing  -> Just p'


actionToWeight :: Action -> Double
actionToWeight WalkLeft  = 1.0
actionToWeight WalkRight = 1.0
actionToWeight JumpLeft  = 1.5
actionToWeight JumpRight = 1.5

-- |Returns a brief representation of the given 'Action'.
ppAction :: Action -> String
ppAction WalkLeft  = "Walk to left"
ppAction WalkRight = "Walk to right"
ppAction JumpLeft  = "Jump to left"
ppAction JumpRight = "Jump to right"