{-# OPTIONS_HADDOCK prune #-}

-- |The constraints that define the relationships between
-- widgets. Each 'Constraints' value can hold 4 'Spring's: one for
-- each edge of the widget. Additionally it can hold 'Spring's for the
-- widget width and the widget height. Since the height and width
-- constraints are dependent on the other constraints, a widget can be
-- over-constraint. In this case (like when all of north, south and
-- height are constraint), the values are adjusted, so that the
-- mathematics still hold true.

module Game.LupupuCube.UI.Constraints
    ( Constraints

    , newConstraints -- private

    , setX
    , setY
    , setWidth
    , setHeight
    , setSouth
    , setEast

    , setNorth
    , setWest

    , unsetX
    , unsetY
    , unsetWidth
    , unsetHeight
    , unsetSouth
    , unsetEast

    , unsetNorth
    , unsetWest

    , getX
    , getY
    , getWidth
    , getHeight
    , getSouth
    , getEast

    , getNorth
    , getWest
    )
    where

import           Control.Monad
import           Control.Monad.ST
import           Data.Maybe
import           Data.STRef
import           Game.LupupuCube.UI.Spring

-- |@'Constraints' s@ is a data type which is parameterised on state
-- @s@.
data Constraints s
    = Constraints {
        cstX      :: STRef s (Maybe (Spring s Horizontal))
      , cstY      :: STRef s (Maybe (Spring s Vertical  ))
      , cstWidth  :: STRef s (Maybe (Spring s Horizontal))
      , cstHeight :: STRef s (Maybe (Spring s Vertical  ))
      , cstEast   :: STRef s (Maybe (Spring s Horizontal))
      , cstSouth  :: STRef s (Maybe (Spring s Vertical  ))
      }


newConstraints :: ST s (Constraints s)
newConstraints = do x <- newSTRef Nothing
                    y <- newSTRef Nothing
                    w <- newSTRef Nothing
                    h <- newSTRef Nothing
                    e <- newSTRef Nothing
                    s <- newSTRef Nothing
                    return $ Constraints x y w h e s


readJust :: STRef s (Maybe a) -> ST s a
readJust = (fromJust `fmap`) . readSTRef

-- |@'setX' cs spr@ sets the west spring of @cs@ to @spr@. If @cs@
-- already has east and width springs, this operation drops the width
-- spring of @cs@.
setX :: Constraints s -> Spring s Horizontal -> ST s ()
setX c s
    = do writeSTRef (cstX c) (Just s)
         e <- readSTRef (cstEast c)
         when (isJust e)
                  $ writeSTRef (cstWidth c) Nothing

-- |Drops the west spring.
unsetX :: Constraints s -> ST s ()
unsetX c = writeSTRef (cstX c) Nothing

-- |@'setY' cs spr@ sets the north spring of @cs@ to @spr@. If @cs@
-- already has south and height springs, this operation drops the
-- height spring of @cs@.
setY :: Constraints s -> Spring s Vertical -> ST s ()
setY c s
    = do writeSTRef (cstY c) (Just s)
         so <- readSTRef (cstSouth c)
         when (isJust so)
                  $ writeSTRef (cstHeight c) Nothing

-- |Drops the north spring.
unsetY :: Constraints s -> ST s ()
unsetY c = writeSTRef (cstY c) Nothing

-- |@'setWidth' cs spr@ sets the width spring of @cs@ to @spr@. If
-- @cs@ already has west and east springs, this operation drops the
-- east spring of @cs@.
setWidth :: Constraints s -> Spring s Horizontal -> ST s ()
setWidth c s
    = do writeSTRef (cstWidth c) (Just s)
         x <- readSTRef (cstX c)
         when (isJust x)
                  $ writeSTRef (cstEast c) Nothing

-- |Drops the width spring.
unsetWidth :: Constraints s -> ST s ()
unsetWidth c = writeSTRef (cstWidth c) Nothing

-- |@'setHeight' cs spr@ sets the height spring of @cs@ to @spr@. If
-- @cs@ already has north and south springs, this operation drops the
-- south spring of @cs@.
setHeight :: Constraints s -> Spring s Vertical -> ST s ()
setHeight c s
    = do writeSTRef (cstHeight c) (Just s)
         y <- readSTRef (cstY c)
         when (isJust y)
                  $ writeSTRef (cstSouth c) Nothing

-- |Drops the height spring.
unsetHeight :: Constraints s -> ST s ()
unsetHeight c = writeSTRef (cstHeight c) Nothing

-- |@'setSouth' cs spr@ sets the south spring of @cs@ to @spr@. If
-- @cs@ already has height and north springs, this operation drops the
-- north spring of @cs@.
setSouth :: Constraints s -> Spring s Vertical -> ST s ()
setSouth c s
    = do writeSTRef (cstSouth c) (Just s)
         h <- readSTRef (cstHeight c)
         when (isJust h)
                  $ writeSTRef (cstY c) Nothing

-- |Drops the south spring.
unsetSouth :: Constraints s -> ST s ()
unsetSouth c = writeSTRef (cstSouth c) Nothing

-- |@'setEast' cs spr@ sets the east spring of @cs@ to @spr@. If @cs@
-- already has width and west springs, this operation drops the west
-- spring of @cs@.
setEast :: Constraints s -> Spring s Horizontal -> ST s ()
setEast c s
    = do writeSTRef (cstEast c) (Just s)
         w <- readSTRef (cstWidth c)
         when (isJust w)
                  $ writeSTRef (cstX c) Nothing

-- |Drops the east spring.
unsetEast :: Constraints s -> ST s ()
unsetEast c = writeSTRef (cstEast c) Nothing

-- |@'getX' cs@ returns the west spring of @cs@. If it doesn't have a
-- west spring, 'getX' returns @east '|-|' width@.
getX :: Constraints s -> ST s (Spring s Horizontal)
getX c
    = do x <- readSTRef (cstX c)
         case x of
           Just sp -> return sp
           Nothing -> readJust (cstEast c) |-| readJust (cstWidth c)

-- |@'getY' cs@ returns the north spring of @cs@. If it doesn't have a
-- north spring, 'getY' returns @south '|-|' height@.
getY :: Constraints s -> ST s (Spring s Vertical)
getY c
    = do y <- readSTRef (cstY c)
         case y of
           Just sp -> return sp
           Nothing -> readJust (cstSouth c) |-| readJust (cstHeight c)

-- |@'getWidth' cs@ returns the width spring of @cs@. If it doesn't
-- have a width spring, 'getWidth' returns @east '|-|' west@.
getWidth :: Constraints s -> ST s (Spring s Horizontal)
getWidth c
    = do w <- readSTRef (cstWidth c)
         case w of
           Just sp -> return sp
           Nothing -> readJust (cstEast c) |-| readJust (cstX c)

-- |@'getHeight' cs@ returns the height spring of @cs@. If it doesn't
-- have a height spring, 'getHeight' returns @south '|-|' north@.
getHeight :: Constraints s -> ST s (Spring s Vertical)
getHeight c
    = do h <- readSTRef (cstHeight c)
         case h of
           Just sp -> return sp
           Nothing -> readJust (cstSouth c) |-| readJust (cstY c)

-- |@'getSouth' cs@ returns the south spring of @cs@. If it doesn't
-- have a south spring, 'getSouth' returns @north '|+|' height@.
getSouth :: Constraints s -> ST s (Spring s Vertical)
getSouth c
    = do s <- readSTRef (cstSouth c)
         case s of
           Just sp -> return sp
           Nothing -> readJust (cstY c) |+| readJust (cstHeight c)

-- |@'getEast' cs@ returns the east spring of @cs@. If it doesn't have
-- a east spring, 'getEast' returns @west '|+|' width@.
getEast :: Constraints s -> ST s (Spring s Horizontal)
getEast c
    = do e <- readSTRef (cstEast c)
         case e of
           Just sp -> return sp
           Nothing -> readJust (cstX c) |+| readJust (cstWidth c)

-- |'setNorth' is a synonym of 'setY'.
setNorth :: Constraints s -> Spring s Vertical -> ST s ()
setNorth = setY

-- |'setWest' is a synonym of 'setX'.
setWest :: Constraints s -> Spring s Horizontal -> ST s ()
setWest = setX

-- |'unsetNorth' is a synonym of 'unsetY'.
unsetNorth :: Constraints s -> ST s ()
unsetNorth = unsetY

-- |'unsetWest' is a synonym of 'unsetX'.
unsetWest :: Constraints s -> ST s ()
unsetWest = unsetX

-- |'getNorth' is a synonym of 'getY'.
getNorth :: Constraints s -> ST s (Spring s Vertical)
getNorth = getY

-- |'getWest' is a synonym of 'getX'.
getWest :: Constraints s -> ST s (Spring s Horizontal)
getWest = getX