{-# OPTIONS_HADDOCK prune #-}

-- |Calculates the space between widget edges, that are layed out by
-- "Game.LupupuCube.UI.SpringLayout".
--
-- A Spring defines a minimum, preferred and maximum distance for each
-- edge (north, east, south, west) of a component.
--
-- However, springs are not static, their actual values are computed
-- at runtime. That means, if a Spring C is defined as the sum of
-- Spring A and Spring B, then the values (min, pref and max) are not
-- calculated at creation of Spring C, but instead always when
-- 'getValue' is applied. So, when Spring A or Spring B changes, this
-- is reflected in Spring C.

module Game.LupupuCube.UI.Spring
    ( Spring(..)
    , SpringT(..)
    , Horizontal
    , Vertical

    -- private
    , unSP1
    , unSP2

    , fixed
    , constant
    , widgetWidth
    , widgetHeight
    , (|+|)
    , (|-|)
    , neg
    , maxSpr
    )
    where

import           Control.Monad.ST
import           Data.STRef
import           UI.HSCurses.Widgets

-- |'Horizontal' is a phantom type whose value is never regarded. It
-- is used by 'Spring' to parameterise it as a horizontal spring.
data Horizontal
-- |'Vertical' is a phantom type whose value is never regarded. It is
-- used by 'Spring' to parameterise it as a vertical spring.
data Vertical

-- |@'Spring' s d@ is a data type which is parameterised on state @s@
-- and direction @d@. @d@ is normally either 'Horizontal' or
-- 'Vertical', and it is actually a phantom type to do merely type
-- checking.
data Spring s d
    = forall i. SpringT s i =>
      SP i -- ^This constructor is exported for those who want to
           -- define their own 'SpringT'. It should not be used by
           -- library users.

-- |@'SpringT' s i@ is a class which is parameterised on state @s@ and
-- implementation @i@. It is exported for spring implementors.
class SpringT s i where
    -- |@'minValue' spr@ returns the minimum value of spring @spr@ in
    -- the 'Control.Monad.ST.ST' monad.
    minValue  :: i -> ST s Int
    -- |@'prefValue' spr@ returns the preferred value of spring @spr@
    -- in the 'Control.Monad.ST.ST' monad.
    prefValue :: i -> ST s Int
    -- |@'maxValue' spr@ returns the maximum value of spring @spr@ in
    -- the 'Control.Monad.ST.ST' monad.
    maxValue  :: i -> ST s Int
    -- |@'getValue' spr@ returns the current value of spring @spr@ in
    -- the 'Control.Monad.ST.ST' monad.
    getValue  :: i -> ST s Int
    -- |@'setValue' spr val@ changes the value of spring @spr@ to
    -- @val@ in the 'Control.Monad.ST.ST' monad.
    setValue  :: i -> Int -> ST s ()


unSP1 :: forall state dir ret.
         (forall impl. SpringT state impl => impl -> ST state ret)
      -> Spring state dir
      -> ST state ret
unSP1 f (SP i) = f i


unSP2 :: forall state dir arg ret.
         (forall impl. SpringT state impl => impl -> arg -> ST state ret)
      -> Spring state dir
      -> arg
      -> ST state ret
unSP2 f (SP i) a = f i a


getStrain :: Spring s d -> ST s Double
getStrain (SP s)
    = do v <- getValue s
         p <- prefValue s
         r <- if v < p
              then shrinkRange (SP s)
              else expandRange (SP s)
         let r' = if r == 0
                  then 1
                  else r
         return $ fromIntegral (v - p) / r'


setStrain :: Double -> Spring s d -> ST s ()
setStrain strain (SP s)
    = do p <- prefValue s
         r <- if strain < 0
              then shrinkRange (SP s)
              else expandRange (SP s)
         setValue s (p + floor (strain * r))


shrinkRange :: Spring s d -> ST s Double
shrinkRange (SP s)
    = do p <- prefValue s
         m <- minValue  s
         return $ fromIntegral (p - m)


expandRange :: Spring s d -> ST s Double
expandRange (SP s)
    = do m <- maxValue  s
         p <- prefValue s
         return $ fromIntegral (m - p)


data Constant s
    = Constant Int Int Int (STRef s Int)

instance SpringT s (Constant s) where
    minValue  (Constant n _ _ _) = return n
    prefValue (Constant _ n _ _) = return n
    maxValue  (Constant _ _ n _) = return n
    getValue  (Constant _ _ _ r) = readSTRef r
    setValue  (Constant a _ b r)
        = \ n -> if n < a || n > b then
                     fail "changing the value of constant Spring is restricted by minValue and maxValue"
                 else
                     writeSTRef r n
                                             
-- |@'fixed' n@ creates a spring whose min, pref and max values are
-- all the same @n@. These kind of springs are struts.
fixed :: Int -> ST s (Spring s d)
fixed n
    = do r <- newSTRef n
         return $ SP $ Constant n n n r

-- |'constant' creates a spring which min, pref and max values are
-- constants.
constant :: Int -- ^The minimum value.
         -> Int -- ^The preferred value.
         -> Int -- ^The maximum value.
         -> ST s (Spring s d)
constant minVal prefVal maxVal
    = do r <- newSTRef prefVal
         return $ SP $ Constant minVal prefVal maxVal r

-- |@'widgetWidth' w@ creates a 'fixed' spring whose value is the
-- width of @'UI.HSCurses.Widgets.minSize' w@.
widgetWidth :: Widget w => w -> ST s (Spring s Horizontal)
widgetWidth = fixed . getWidth . minSize

-- |@'widgetHeight' w@ creates a 'fixed' spring whose value is the
-- height of @'UI.HSCurses.Widgets.minSize' w@.
widgetHeight :: Widget w => w -> ST s (Spring s Vertical)
widgetHeight = fixed . getHeight . minSize


data Add s d = Add (Spring s d) (Spring s d)

instance SpringT s (Add s d) where
    minValue    (Add (SP a) (SP b)) = do { m <- minValue  a; n <- minValue  b; return (m + n) }
    prefValue   (Add (SP a) (SP b)) = do { m <- prefValue a; n <- prefValue b; return (m + n) }
    maxValue    (Add (SP a) (SP b)) = do { m <- maxValue  a; n <- maxValue  b; return (m + n) }
    getValue    (Add (SP a) (SP b)) = do { m <- getValue  a; n <- getValue  b; return (m + n) }
    setValue sp@(Add (SP a) (SP b))
        = \ n -> do strain <- getStrain (SP sp)
                    setStrain strain (SP a)
                    remainder <- return . (n -) =<< getValue a
                    setValue b remainder

-- |@a '|+|' b@ creates a spring which is always the sum of @a@ and
-- @b@.
(|+|) :: ST s (Spring s d) -> ST s (Spring s d) -> ST s (Spring s d)
a |+| b = do a' <- a
             b' <- b
             return (SP (Add a' b'))


data Negate s d = Negate (Spring s d)

instance SpringT s (Negate s d) where
    minValue  (Negate (SP a)) = minValue  a >>= return . negate
    prefValue (Negate (SP a)) = prefValue a >>= return . negate
    maxValue  (Negate (SP a)) = maxValue  a >>= return . negate
    getValue  (Negate (SP a)) = getValue  a >>= return . negate
    setValue  (Negate (SP a)) = setValue  a . negate

-- |@'neg' a@ creates a spring which is always the negation of @a@.
neg :: ST s (Spring s d) -> ST s (Spring s d)
neg = ((SP . Negate) `fmap`)

-- |@a '|-|' b@ creates a spring which is always the subtraction @a -
-- b@ (i.e. @a '|+|' 'neg' b@).
(|-|) :: ST s (Spring s d) -> ST s (Spring s d) -> ST s (Spring s d)
a |-| b = a |+| neg b


data Max s d = Max (Spring s d) (Spring s d)

instance SpringT s (Max s d) where
    minValue  (Max (SP a) (SP b)) = do { m <- minValue  a; n <- minValue  b; return (m `max` n) }
    prefValue (Max (SP a) (SP b)) = do { m <- prefValue a; n <- prefValue b; return (m `max` n) }
    maxValue  (Max (SP a) (SP b)) = do { m <- maxValue  a; n <- maxValue  b; return (m `max` n) }
    getValue  (Max (SP a) (SP b)) = do { m <- getValue  a; n <- getValue  b; return (m `max` n) }
    setValue  (Max (SP a) (SP b))
        = \ n -> do p1 <- prefValue a
                    p2 <- prefValue b
                    if p1 < p2 then
                        do setValue a (n `min` p1)
                           setValue b n
                      else
                        do setValue a n
                           setValue b (n `min` p2)

-- |@'maxSpr' a b@ creates a spring which always has the maximum
-- values among @a@ and @b@.
maxSpr :: ST s (Spring s d) -> ST s (Spring s d) -> ST s (Spring s d)
a `maxSpr` b = do a' <- a
                  b' <- b
                  return (SP (Max a' b'))
