-- |A very flexible layout manager. Widgets are laid out by defining
-- the relationships between them. The relationships are expressed as
-- 'Spring's. You can attach a 'Spring' for each edge of a widget and
-- link it to an edge of a different widget. For example, you can say,
-- the northern edge of widget A should be attached to the southern
-- edge of widget B, and the space between them should be something
-- between x and y pixels, and preferably z pixels.

module Game.LupupuCube.UI.SpringLayout
    ( SpringLayoutWidget
    , Adder

    , createSpringLayout

    , module Game.LupupuCube.UI.Constraints
    , module Game.LupupuCube.UI.Spring
    )
    where

import           Control.Monad.ST
import           Data.STRef
import           Game.LupupuCube.UI.Constraints
import           Game.LupupuCube.UI.Spring
import qualified UI.HSCurses.Widgets as W
import           UI.HSCurses.Widgets (Widget)


data Component = forall w. Widget w => Component w

-- |This is a widget containing number of other widgets and their
-- relationships.
data SpringLayoutWidget = SpringLayoutWidget (forall s. ST s (SpringLayoutContent s))

data SpringLayoutContent s
    = SpringLayoutContent {
        slcConstraints :: Constraints s
      , slcComponents  :: STRef s [(Component, Constraints s)]
      }

instance Widget SpringLayoutWidget where
    draw pos sz hint (SpringLayoutWidget widget)
        = stToIO $
          do slc <- widget
             let c = slcConstraints slc
             flip (unSP2 setValue) 0 =<< getX c
             flip (unSP2 setValue) 0 =<< getY c
             flip (unSP2 setValue) (W.getWidth  sz) =<< getWidth  c
             flip (unSP2 setValue) (W.getHeight sz) =<< getHeight c
             mapM_ draw' =<< readSTRef (slcComponents slc)
        where
          draw' (Component comp, c)
              = do x <- unSP1 getValue =<< getX c
                   y <- unSP1 getValue =<< getY c
                   w <- unSP1 getValue =<< getWidth  c
                   h <- unSP1 getValue =<< getHeight c
                   unsafeIOToST $ W.draw (y + W.getYPos pos, x + W.getXPos pos) (h, w) hint comp

    minSize (SpringLayoutWidget widget)
        = runST $
          do slc <- widget
             let c = slcConstraints slc
             minW <- unSP1 minValue =<< getWidth  c
             minH <- unSP1 minValue =<< getHeight c
             return (minH, minW)

-- |@adder w@ inserts the widget @w@ into the 'SpringLayoutWidget' of
-- the current context, and then it returns a new 'Constraints' for
-- the widget @w@ in the 'Control.Monad.ST.ST' monad.
type Adder s = forall w. Widget w => w -> ST s (Constraints s)

-- |Oh no, this function is malciously designed to destroy your
-- sanity. What's the heck is this?
--
-- 'createSpringLayout' creates a 'SpringLayoutWidget' using a given
-- function which takes the constraints of the 'SpringLayoutWidget'
-- itself, and an 'Adder' function, and returns @()@ in the
-- 'Control.Monad.ST.ST' monad.
--
-- The initial 'Constraints' for the 'SpringLayoutWidget' itself is as
-- follows:
--
--   [west] 'fixed' 0
--
--   [north] 'fixed' 0
--
--   [width] 'constant' 0 0 'Prelude.maxBound'
--
--   [height] 'constant' 0 0 'Prelude.maxBound'
--
-- The initial 'Constraints' for each new widgets are as follows:
--
--   [west] 'fixed' 0
--
--   [north] 'fixed' 0
--
--   [width] 'widgetWidth' w
--
--   [height] 'widgetHeight' w
--
-- Example:
--
-- >  mkMainWidget :: SpringLayoutWidget
-- >  mkMainWidget = createSpringLayout $ \ frame add ->
-- >                 do widget1 <- add mkWidget1
-- >                    widget2 <- add mkWidget2
-- >
-- >                    setEast widget2 =<< getEast frame
-- >                    setX    widget2 =<< getEast widget1
-- >
-- >  mkWidget1 :: SomeWidget
-- >  mkWidget2 :: SomeWidget
createSpringLayout :: (forall s. Constraints s -> Adder s -> ST s ())
                   -> SpringLayoutWidget
createSpringLayout f
    = SpringLayoutWidget $ do slc <- newSLC
                              f (slcConstraints slc) (addComponent slc)
                              return slc


newSLC :: ST s (SpringLayoutContent s)
newSLC = do c <- newConstraints
            setX      c  =<< fixed 0
            setY      c  =<< fixed 0
            setWidth  c  =<< constant 0 0 maxBound
            setHeight c  =<< constant 0 0 maxBound
            cs <- newSTRef []
            return $ SpringLayoutContent c cs
    

addComponent :: Widget w => SpringLayoutContent s -> w -> ST s (Constraints s)
addComponent slc widget
    = do c <- newConstraints
         setX      c  =<< fixed 0
         setY      c  =<< fixed 0
         setWidth  c  =<< widgetWidth  widget
         setHeight c  =<< widgetHeight widget
         modifySTRef (slcComponents slc) ((Component widget, c) :)
         return c
