-- |Read stages from XML files, and write solutions to XML files.

module Game.LupupuCube.XML
    ( Statistics(..)

    , readStage
    , writeSolution
    )
    where

import           Data.Time
import           Game.LupupuCube.Phase
import qualified Game.LupupuCube.Schema.Stage as SS
import           Game.LupupuCube.Stage
import           Game.LupupuCube.Solver
import           Game.LupupuCube.XmlIO
import           Text.XML.HXT.Arrow
import           Text.XML.HXT.RelaxNG

-- |'Statistics' for a process of stage solving.
data Statistics
    = Statistics {
        -- |The number of seconds spent to find solutions or to prove
        -- unexistence of any.
        statElapsed      :: !NominalDiffTime

        -- |The number of states examined to find solutions or to
        -- prove unexistence of any.
      , statTestedStates :: !Int
      }
    deriving (Eq, Show)

instance XmlWrite Statistics where
    xmlWrite = eelem "statistics"
               += ( eelem "elapsedSeconds"
                    += (arr (showNDT . statElapsed) >>> mkText)
                  )
               += ( eelem "testedStates"
                    += (arr (show . statTestedStates) >>> mkText)
                  )
               += (arr sps >>> mkCmt)
        where
          showNDT :: NominalDiffTime -> String
          showNDT ndt = let d = fromRational $ toRational ndt :: Double
                        in
                          show d

          sps :: Statistics -> String
          sps s = let sec = fromRational $ toRational $ statElapsed s :: Double
                      n   = fromIntegral $ statTestedStates s         :: Double
                  in
                    "States per second: " ++ show (n / sec)

-- |Validates and reads a 'Stage' from 'XmlTree'.
--
-- [Example of input XML] See example/simple.xml
--
-- [The RelaxNG schema] See schemas/lupu-stage-1.0.rng
readStage :: IOSArrow XmlTree Stage
readStage = perform ( validateDocumentWithRelax SS.schema
                      >>>
                      getErrors
                      >>> 
                      getChildren >>> getAttrValue "desc"
                      >>>
                      arr ("RelaxNG validation: " ++)
                      >>> 
                      mkError c_err
                      >>>
                      filterErrorMsg
                    )
            >>>
            getChildren
            >>>
            xmlRead

-- |Writes a 'Stage', 'Solution' and 'Statistics' to 'XmlTree'.
--
-- [Example of output XML] Run @roro example/simple.xml@ and see what
-- appears to the stdout.
writeSolution :: IOSArrow (Stage, (Maybe Solution, Maybe Statistics)) XmlTree
writeSolution
    = eelem "/"
      += ( eelem "lupuSolution"
           += sattr "xmlns" "http://cielonegro.org/schema/RoRo/Solution/1.0"
           += ( proc (st, _) -> xmlWrite -< st )
           += ( proc (st, (solM, _)) ->
                do size <- arr stSize    -< st
                   sol  <- dissolveMaybe -< solM
                   mkSol -< (size, sol)
              )
           += ( proc (_, (_, stat)) -> dissolveMaybe >>> xmlWrite -< stat )
         )


mkSol :: (ArrowXml a, ArrowChoice a) => a (Size, Solution) XmlTree
mkSol = eelem "solution"
        += ( proc (size, sol) ->
             do ns          <- constA [1..]         -< ()
                (n, (a, p)) <- unlistA <<< arr2 zip -< (ns, sol)
                qs          <- mkQS                 -< (size, (n, (a, p)))
                ae          <- mkAE                 -< a

                unlistA -< [qs, ae]
           )


mkQS :: ArrowXml a => a (Size, (Int, (Action, Phase))) XmlTree
mkQS = proc (size, (n, (a, p))) ->
       let str = show n ++ ": " ++ ppAction a ++ "\n" ++
                 ppPhase size p
       in
         mkCmt -< str
       

mkAE :: (ArrowXml a, ArrowChoice a) => a Action XmlTree
mkAE = proc a ->
       case a of
         WalkLeft  -> eelem "walkLeft"  -< ()
         WalkRight -> eelem "walkRight" -< ()
         JumpLeft  -> eelem "jumpLeft"  -< ()
         JumpRight -> eelem "jumpRight" -< ()


dissolveMaybe :: ArrowList a => a (Maybe b) b
dissolveMaybe = arr (maybe [] (:[])) >>> unlistA
