import           Control.Concurrent
import           Control.Exception
import           Control.Monad
import           Data.IORef
import           Data.List
import           Data.Time
import           Game.LupupuCube
import           Game.LupupuCube.UI
import           Game.LupupuCube.UI.PhaseWidget
import           System.Console.GetOpt
import           System.Environment
import           System.Exit
import           Text.XML.HXT.Arrow hiding (when, attr, loop, second)
import           UI.HSCurses.Curses
import           UI.HSCurses.CursesHelper as CH
import           UI.HSCurses.Widgets hiding (getWidth, getHeight)


data CmdOpt
    = OptOutput FilePath
    | OptQuiet
    | OptHelp
    deriving (Eq, Show)


options :: [OptDescr CmdOpt]
options = [ Option ['o'] ["output"]
                   (ReqArg OptOutput "FILE")
                   "File to write the result. (default: stdout)"

          , Option ['q'] ["quiet"]
                   (NoArg OptQuiet)
                   "Be quiet."

          , Option ['h'] ["help"]
                   (NoArg OptHelp)
                   "Print this message."
          ]


printUsage :: IO ()
printUsage = do putStrLn "Usage:"
                putStrLn "  roro [URL] [OPTIONS...]"
                putStrLn ""
                putStrLn "Input URL:"
                putStrLn "  file path: read from the file"
                putStrLn "  \"-\"      : read from stdin"
                putStrLn "  (omitted): ditto"
                putStrLn ""
                putStr $ usageInfo "Options:" options


main :: IO ()
main = do (opts, nonOpts, errors) <- return . getOpt Permute options =<< getArgs

          unless (null errors)
                   $ do mapM_ putStr errors
                        exitWith (ExitFailure 1)

          when (any (\ x -> x == OptHelp) opts)
                   $ do printUsage
                        exitWith ExitSuccess

          inFile <- case nonOpts of
                      []  -> return "stdin:"
                      [f] -> return f
                      _   -> do printUsage
                                exitWith (ExitFailure 1)

          outFile <- case filter (\ x -> case x of
                                           OptOutput _ -> True
                                           _           -> False) opts of
                       []            -> return "-"
                       [OptOutput o] -> return o
                       _             -> do printUsage
                                           exitWith (ExitFailure 1)


          [stage] <- runX ( setErrorMsgHandler False fail
                            >>>
                            readDocument [ (a_validate        , v_0 )
                                         , (a_check_namespaces, v_1 )
                                         , (a_encoding        , utf8)
                                         ] inFile
                            >>>
                            readStage
                          )

          let solver = if any (\ x -> x == OptQuiet) opts
                       then solveQuietly
                       else solveNoisily

          (sol, stat) <- solver stage

          runX ( setErrorMsgHandler False fail
                 >>>
                 constA (stage, (sol, Just stat))
                 >>>
                 writeSolution
                 >>>
                 writeDocument [ (a_indent, v_1) ] outFile
               )
          return ()


type Clock = IO NominalDiffTime

mkClock :: IO Clock
mkClock = do origin <- getCurrentTime
             return $ do now <- getCurrentTime
                         return (now `diffUTCTime` origin)


solveQuietly :: Stage -> IO (Maybe Solution, Statistics)
solveQuietly st
    = do ctr <- newIORef undefined

         clock   <- mkClock
         sol     <- solveM st (count ctr)
         elapsed <- clock
         nStates <- readIORef ctr

         let stat = Statistics {
                      statElapsed      = elapsed
                    , statTestedStates = nStates
                    }
         return (sol, stat)
    where
      count :: IORef Int -> LupuGraph -> IO ()
      count ctr lg
          | isFinished lg = writeIORef ctr (numOfStates lg)
          | otherwise     = return ()

{-
     0        10        20        30        40        50        60        70        80
 0 __|....i....|....i....|....i....|....i....|....i....|....i....|....i....|....i....|.__  0
     .####............####..........                                                  .
     .##..##..........##..##........     The LupupuCube Solver written in Haskell     .
     .##..##...%%%%...##..##...%%%%.                                                  .
     .#####...%%..%%..#####...%%..%%             (PD) Programmed by PHO               .
    --##..##..%%..%%..##..##..%%..%%             http://cielonegro.org/               --
     .##..##...%%%%...##..##...%%%%.                                                  .
     .Solving_stage:_foo_bar_baz______________________________________________________.
     .Tested states   Pending states   Current moves   Time spent   States/Sec        .
     .       195828           395029              12   1h 11m 11s          512        .
10 ---                                                                                --- 10
     .                                                                                .
     .                                                                                .
     .                                ............                                    .
     .                                ............                                    .
    --                                ............                                    --
     .                                ............                                    .
     .                                ............                                    .
     .                                ............                                    .
     .                                ............                                    .
20 ---                                ............                                    --- 20
     .                                                                                .
     .                                                                                .
     .                                                                                .
     .                                  Type `C-c' or`q' to terminate the computation..
     |''''i''''|''''i''''|''''i''''|''''i''''|''''i''''|''''i''''|''''i''''|''''i''''|'
     0        10        20        30        40        50        60        70        80
-}
solveNoisily :: Stage -> IO (Maybe Solution, Statistics)
solveNoisily st = bracket_ CH.start CH.end worker
    where
      worker :: IO (Maybe Solution, Statistics)
      worker = do kh <- spawnKeyHandler =<< myThreadId

                  cursSet CursorInvisible
                  erase

                  ctr <- newIORef undefined

                  clock   <- mkClock
                  sol     <- do lastUpdated <- newIORef 0
                                solveM st $ \ lg ->
                                    do lastUpd <- readIORef lastUpdated
                                       elapsed <- clock
                                       when (elapsed - lastUpd >= updatesPerSec)
                                                $ do updateScreen elapsed lg
                                                     writeIORef lastUpdated elapsed
                                       count ctr lg
                  elapsed <- clock
                  nStates <- readIORef ctr

                  let stat = Statistics {
                               statElapsed      = elapsed
                             , statTestedStates = nStates
                             }

                  killThread kh
                  return (sol, stat)

      updatesPerSec = 0.1

      spawnKeyHandler :: ThreadId -> IO ThreadId
      spawnKeyHandler mainTid = forkIO loop
          where
            loop :: IO ()
            loop = do k <- getKey refresh
                      case k of
                        KeyChar 'q'
                            -> killThread mainTid
                        KeyChar '\x03' -- C-c
                            -> killThread mainTid
                        _   -> loop

      updateScreen :: NominalDiffTime -> LupuGraph -> IO ()
      updateScreen elapsed lg = redrawMain mw
          where
            mw = mkMainWidget st elapsed lg

      count :: IORef Int -> LupuGraph -> IO ()
      count ctr lg
          | isFinished lg = writeIORef ctr (numOfStates lg)
          | otherwise     = return ()


redrawMain :: SpringLayoutWidget -> IO ()
redrawMain w = do sz <- scrSize
                  draw (0, 0) sz DHNormal w
                  refresh



mkMainWidget :: Stage -> NominalDiffTime -> LupuGraph -> SpringLayoutWidget
mkMainWidget stage elapsed lg
    = createSpringLayout $ \ frame add ->
      do logo <- add roroLogo

         progTitle <- add programTitle
         setEast   progTitle =<< getEast   frame
         setX      progTitle =<< getEast   logo
         setHeight progTitle =<< getHeight logo

         stageTitle <- add (mkStageTitle stage)
         setEast stageTitle =<< getEast frame
         setX    stageTitle =<< getX    frame
         setY    stageTitle =<< getSouth logo

         testedStates <- add (mkTestedStates lg)
         setX testedStates =<< getX frame
         setY testedStates =<< getSouth stageTitle

         pendingStates <- add (mkPendingStates lg)
         setX pendingStates =<< getEast testedStates |+| constant 1 3 3
         setY pendingStates =<< getSouth stageTitle

         currentMoves <- add (mkCurrentMoves lg)
         setX currentMoves =<< getEast pendingStates |+| constant 1 3 3
         setY currentMoves =<< getSouth stageTitle

         timeSpent <- add (mkTimeSpent elapsed)
         setX timeSpent =<< getEast currentMoves |+| constant 1 3 3
         setY timeSpent =<< getSouth stageTitle

         statesPerSec <- add (mkStatesPerSec lg elapsed)
         setX statesPerSec =<< getEast timeSpent |+| constant 1 3 3
         setY statesPerSec =<< getSouth stageTitle

         msgLine <- add messageLine
         -- FIXME: This seems to expose a possible bug in the ncurses.
         -- FIXME: setEast msgLine =<< getEast frame
         setEast  msgLine =<< getEast frame |-| fixed 1
         setSouth msgLine =<< getSouth frame

         nextPhase <- add (mkNextPhase lg)
         setEast  nextPhase =<< getEast  frame
         setX     nextPhase =<< getX     frame
         setSouth nextPhase =<< getY     msgLine
         setY     nextPhase =<< getSouth testedStates


mkStageTitle :: Stage -> TextWidget
mkStageTitle st = newTextWidget opts str
    where
      opts = defaultTWOptions {
               twopt_style  = mkDrawingStyle (mkCursesStyle [Reverse])
             , twopt_halign = AlignCenter
             }
      str = "Solving stage: " ++ stTitle st


mkTestedStates :: LupuGraph -> TextWidget
mkTestedStates lg = newTextWidget opts str
    where
      opts = defaultTWOptions {
               twopt_halign = AlignRight
             }
      str = "Tested states\n" ++ show (numOfStates lg)


mkPendingStates :: LupuGraph -> TextWidget
mkPendingStates lg = newTextWidget opts str
    where
      opts = defaultTWOptions {
               twopt_halign = AlignRight
             }
      str  = "Pending states\n" ++ show (numOfPendings lg)


mkCurrentMoves :: LupuGraph -> TextWidget
mkCurrentMoves lg = newTextWidget opts str
    where
      opts = defaultTWOptions {
               twopt_halign = AlignRight
             }
      str  = "Current moves\n" ++ maybe "-" (show . depth) (nextPending lg)
      depth :: Pending -> Depth
      depth (PEN _ n _) = n


mkTimeSpent :: NominalDiffTime -> TextWidget
mkTimeSpent elapsed = newTextWidget opts str
    where
      hour, minute, second :: Int
      hour   = floor (elapsed / 60 / 60)
      minute = floor ((elapsed - (fromIntegral $ hour * 60 * 60)) / 60)
      second = floor (elapsed - (fromIntegral $ hour * 60 * 60) - (fromIntegral $ minute * 60))

      opts = defaultTWOptions {
               twopt_halign = AlignRight
             }
      str  = "Time spent\n" ++
             show hour ++ "h " ++ show minute ++ "m " ++ show second ++ "s"


mkStatesPerSec :: LupuGraph -> NominalDiffTime -> TextWidget
mkStatesPerSec lg elapsed = newTextWidget opts str
    where
      sps :: Int
      sps = floor (fromIntegral (numOfStates lg) / elapsed)

      opts = defaultTWOptions {
               twopt_halign = AlignRight
             }
      str  = "States/Sec\n" ++ show sps


mkNextPhase :: LupuGraph -> SpacingLayoutWidget
mkNextPhase lg
    = case nextPending lg of
        Nothing
            -> mkSpacingLayout (EmptyWidget (1, 1))
        Just (PEN _ _ p)
            -> mkSpacingLayout (PhaseWidget (lgSize lg) p)


programTitle :: SpacingLayoutWidget
programTitle = mkSpacingLayout (newTextWidget opts str)
    where
      opts = defaultTWOptions {
               twopt_halign = AlignCenter
             }
      str = "The LupupuCube Solver written in Haskell\n" ++
            "\n" ++
            "(PD) Programmed by PHO\n" ++
            "http://cielonegro.org/"


messageLine :: TextWidget
messageLine = newTextWidget opts str
    where
      opts = defaultTWOptions {
               twopt_style = mkDrawingStyle (mkCursesStyle [Bold])
             }
      str = "Type `Ctrl-C' or `q' to terminate the program."


roroLogo :: LogoWidget
roroLogo = LogoWidget

data LogoWidget = LogoWidget
instance Widget LogoWidget where
    draw (y, x) _ _ _
        = withStyle defaultCursesStyle $
          do mvWAddStr stdScr y     x "####............####.........."
             mvWAddStr stdScr (y+1) x "##..##..........##..##........"
             mvWAddStr stdScr (y+2) x "##..##...%%%%...##..##...%%%%."
             mvWAddStr stdScr (y+3) x "#####...%%..%%..#####...%%..%%"
             mvWAddStr stdScr (y+4) x "##..##..%%..%%..##..##..%%..%%"
             setStyle (mkCursesStyle [Underline])
             mvWAddStr stdScr (y+5) x "##..##...%%%%...##..##...%%%%."

    minSize _ = (6, 30)
